use leptos::*;
use leptos_router::*;
use serde::{Deserialize, Serialize};
use uuid::Uuid;

use super::format_errors;

#[component]
pub fn ViewingList(series_id: Uuid) -> impl IntoView {
    let add = create_server_action::<AddViewing>();
    let viewings = create_resource(
        move || (add.version().get(), series_id),
        |(_, id)| async move { get_viewings(id).await },
    );
    view! {
        <h3 class="mb-3">Viewing History</h3>
        <Suspense fallback=|| view!{ <div>Loading...</div> }>
        {viewings().map(format_viewings)}
        </Suspense>
        <ActionForm action=add>
        <input type="hidden" name="series_id" value=series_id.to_string() />
        <div class="row g-3">
            <div class="col col-4 align-middle">
                <input class="form-control btn btn-primary" type="submit" value="Watch Episode" />
            </div>
            <div class="col">
                <div class="input-group mb-3">
                    <label class="input-group-text" for="season">Season</label>
                    <input class="form-control" name="season" type="number" value="1" />
                    <label class="input-group-text" for="episode">Episode</label>
                    <input class="form-control" name="episode" type="number" value="1" />
                </div>
                <div>
                    <label class="form-label" for="comment">Comment</label>
                    <textarea class="form-control" name="comment" />
                </div>
            </div>
        </div>
        </ActionForm>
    }
}

fn format_viewings(
    result: Result<Vec<Viewing>, ServerFnError>,
) -> Result<impl IntoView, ServerFnError> {
    Ok(view! {
        <ErrorBoundary fallback=format_errors>
        {result.map(|viewings| {
            let empty = viewings.is_empty();
            view! {
                <Show when=move || empty fallback=|| ()>
                    <div class="mb-3">No viewings found.</div>
                </Show>
                <ul class="list-group mb-3">
                {viewings.into_iter().map(|viewing| {
                    view! {
                        <li class="list-group-item">
                            Season {viewing.season},
                            Episode {viewing.episode}
                            on {viewing.created}
                        <Suspense fallback=|| ()>
                        {viewing.comment.as_ref()}
                        </Suspense>
                        </li>
                    }
                }).collect_view()}
                </ul>
            }
        })}
        </ErrorBoundary>
    })
}

#[derive(Serialize, Deserialize, Debug, Clone)]
pub struct Viewing {
    pub id: Uuid,
    pub created: String,
    pub updated: String,
    pub season: i16,
    pub episode: i16,
    pub comment: Option<String>,
}

#[cfg(feature = "ssr")]
mod ssr {
    use super::Viewing;
    use sqlx::{
        postgres::PgRow,
        prelude::*,
        types::chrono::{DateTime, Utc},
    };

    impl<'r> FromRow<'r, PgRow> for Viewing {
        fn from_row(row: &'r PgRow) -> Result<Self, sqlx::Error> {
            let id = row.try_get("id")?;
            let created: DateTime<Utc> = row.try_get("created")?;
            let created = created.format("%Y-%m-%d").to_string();
            let updated: DateTime<Utc> = row.try_get("updated")?;
            let updated = updated.format("%Y-%m-%d").to_string();
            let season = row.try_get("season")?;
            let episode = row.try_get("episode")?;
            let comment = row.try_get("comment")?;

            Ok(Viewing {
                id,
                created,
                updated,
                season,
                episode,
                comment,
            })
        }
    }
}

#[server(AddViewing)]
pub async fn add_viewing(
    series_id: Uuid,
    episode: i32,
    season: i32,
    comment: Option<String>,
) -> Result<(), ServerFnError> {
    use std::sync::OnceLock;

    use sqlx::{Pool, Postgres};

    use crate::db::load_statement;

    static SQL: OnceLock<String> = OnceLock::new();

    let pool = expect_context::<Pool<Postgres>>();
    let s = SQL.get_or_init(|| load_statement("insert_viewing.sql"));
    sqlx::query(s)
        .persistent(true)
        .bind(series_id)
        .bind(season)
        .bind(episode)
        .bind(comment)
        .execute(&pool)
        .await?;

    Ok(())
}

#[server(GetViewings)]
pub async fn get_viewings(series_id: Uuid) -> Result<Vec<Viewing>, ServerFnError> {
    use std::sync::OnceLock;

    use sqlx::{Pool, Postgres};

    use crate::db::load_statement;

    static SQL: OnceLock<String> = OnceLock::new();

    let pool = expect_context::<Pool<Postgres>>();
    let s = SQL.get_or_init(|| load_statement("select_viewings.sql"));
    Ok(sqlx::query_as::<_, Viewing>(s)
        .persistent(true)
        .bind(series_id)
        .fetch_all(&pool)
        .await?)
}
